from __future__ import annotations

from itertools import chain
from typing import Callable, Iterator, Mapping, TypeVar

from sm.namespaces.namespace import KnowledgeGraphNamespace

V = TypeVar("V")


class KGMapping(Mapping[str, V]):
    def __init__(
        self,
        main: Mapping[str, V],
        default: Mapping[str, V],
        uri_to_id: Callable[[str], str],
    ):
        self.main = main
        self.default = default
        self.uri_to_id = uri_to_id

    def __getitem__(self, key: str):
        if key in self.main:
            return self.main[key]
        return self.default[key]

    def __iter__(self) -> Iterator[str]:
        return chain(iter(self.main), iter(self.default))

    def values(self) -> Iterator[V]:
        return chain(self.main.values(), self.default.values())

    def __len__(self) -> int:
        return len(self.main) + len(self.default)

    def __contains__(self, key: str) -> bool:
        return key in self.main or key in self.default

    def get(self, key: str, default=None):
        if key in self.main:
            return self.main[key]
        return self.default.get(key, default)

    def get_by_uri(self, uri: str, default=None):
        return self.get(self.uri_to_id(uri), default)



from __future__ import annotations

from dataclasses import asdict
from typing import List, Optional

import sm.outputs.semantic_model as O
from dependency_injector.wiring import Provide, inject
from playhouse.shortcuts import model_to_dict
from sm.misc.funcs import assert_not_null

from sand.helpers.namespace import NamespaceService
from sand.models import SemanticModel
from sand.models.entity import Entity
from sand.models.ontology import OntClass, OntProperty


def serialize_property(obj: OntProperty):
    return {
        "id": obj.id,
        "uri": obj.uri,
        "label": obj.label,
        "readable_label": obj.readable_label,
        "aliases": obj.aliases,
        "description": obj.description,
        "datatype": obj.datatype,
        "parents": obj.parents,
        "ancestors": list(obj.ancestors.keys()),
    }


def serialize_class(obj: OntClass):
    return {
        "id": obj.id,
        "uri": obj.uri,
        "label": obj.label,
        "readable_label": obj.readable_label,
        "aliases": obj.aliases,
        "description": obj.description,
        "parents": obj.parents,
        "ancestors": list(obj.ancestors.keys()),
    }


def serialize_entity(obj: Entity):
    return {
        "id": obj.id,
        "uri": obj.uri,
        "readable_label": obj.readable_label,
        "label": obj.label.to_dict(),
        "aliases": obj.aliases.to_dict(),
        "description": obj.description.to_dict(),
        "properties": {
            prop: [
                {
                    "value": asdict(stmt.value),
                    "qualifiers": {
                        qid: [asdict(qval) for qval in qvals]
                        for qid, qvals in stmt.qualifiers.items()
                    },
                    "qualifiers_order": stmt.qualifiers_order,
                }
                for stmt in stmts
            ]
            for prop, stmts in obj.properties.items()
        },
    }


@inject
def serialize_graph(
    sm: O.SemanticModel,
    columns: Optional[List[str]] = None,
    namespace: NamespaceService = Provide["namespace"],
    # ontclass_ar: OntClassAR = Provide["classes"],
):
    nodes = []
    for n in sm.iter_nodes():
        if isinstance(n, O.ClassNode):
            nodes.append(
                {
                    "id": str(n.id),
                    "uri": n.abs_uri,
                    "label": n.label,
                    "approximation": n.approximation,
                    "type": "class_node",
                }
            )
        elif isinstance(n, O.DataNode):
            nodes.append(
                {
                    "id": str(n.id),
                    "label": columns[n.col_index] if columns is not None else n.label,
                    "type": "data_node",
                    "column_index": n.col_index,
                }
            )
        else:
            if n.datatype == O.LiteralNodeDataType.Entity:
                value = {
                    "id": namespace.uri_to_id(n.value),
                    "uri": n.value,
                    "type": n.datatype.value,
                }
            else:
                value = {"type": n.datatype.value, "value": n.value}
            nodes.append(
                {
                    "id": str(n.id),
                    "value": value,
                    "label": n.label,
                    "type": "literal_node",
                    "is_in_context": n.is_in_context,
                }
            )

    edges = [
        {
            "source": str(e.source),
            "target": str(e.target),
            "uri": e.abs_uri,
            "label": e.label,
            # "label": e.readable_label
            # or (
            #     tmp.readable_label
            #     if (tmp := self.ontprop_ar.get_by_uri(e.abs_uri)) is not None
            #     else e.rel_uri
            # ),
            "approximation": e.approximation,
        }
        for e in sm.iter_edges()
    ]

    return dict(nodes=nodes, edges=edges)


def batch_serialize_sms(sms: List[SemanticModel]):
    output = []
    for sm in sms:
        r = model_to_dict(sm, recurse=False)
        if r["data"] is not None:
            # columns = None if sm.table_id is None else tbls[sm.table_id].columns
            columns = None
            r["data"] = serialize_graph(r["data"], columns)  # type: ignore
        output.append(r)
    return output



from dataclasses import dataclass, field
from typing import Dict, List, Literal, Mapping

from dependency_injector.wiring import Provide, inject
from rdflib import RDF, RDFS
from sm.misc.funcs import import_func

from sand.config import AppConfig
from sand.helpers.mapping_utils import KGMapping
from sand.helpers.namespace import NamespaceService


@dataclass
class OntClass:
    id: str
    uri: str
    label: str
    aliases: List[str]
    description: str
    parents: List[str]
    ancestors: Dict[str, int] = field(default_factory=dict)

    @property
    def readable_label(self):
        return self.label


OntPropertyDataType = Literal[
    "monolingualtext",
    "url",
    "entity",
    "datetime",
    "integer-number",
    "decimal-number",
    "string",
    "globe-coordinate",
    "unknown",  # don't know the type yet
]


@dataclass
class OntProperty:
    id: str
    uri: str
    label: str
    datatype: OntPropertyDataType
    aliases: List[str]
    description: str
    parents: List[str]
    ancestors: Dict[str, int] = field(default_factory=dict)

    @property
    def readable_label(self):
        return self.label


def semweb_default_props():
    return {
        "rdfs:label": OntProperty(
            id="rdfs:label",
            uri=str(RDFS.label),
            label="rdfs:label",
            aliases=[],
            datatype="string",
            description="Provides a human-readable version of a resource's name.",
            parents=[],
        ),
        "rdf:type": OntProperty(
            id="rdf:type",
            uri=str(RDF.type),
            label="rdf:type",
            aliases=[],
            datatype="entity",
            description="Is used to state that a resource is an instance of a class",
            parents=[],
        ),
    }


@inject
def get_default_properties(cfg: AppConfig = Provide["appcfg"]):
    mapping = import_func(cfg.property.default)()
    return mapping


@inject
def get_default_classes(cfg: AppConfig = Provide["appcfg"]):
    mapping = import_func(cfg.clazz.default)()
    return mapping


class OntClassAR(KGMapping[OntClass]):
    @staticmethod
    @inject
    def init(
        appcfg: AppConfig = Provide["appcfg"],
        namespace: NamespaceService = Provide["namespace"],
        default_classes: Mapping[str, OntClass] = Provide["default_classes"],
    ):
        cfg = appcfg.clazz
        func = import_func(cfg.constructor)
        return OntClassAR(func(**cfg.args), default_classes, namespace.uri_to_id)


class OntPropertyAR(KGMapping[OntProperty]):
    @staticmethod
    @inject
    def init(
        appcfg: AppConfig = Provide["appcfg"],
        namespace: NamespaceService = Provide["namespace"],
        default_properties: Mapping[str, OntProperty] = Provide["default_properties"],
    ):
        func = import_func(appcfg.property.constructor)
        return OntPropertyAR(
            func(**appcfg.property.args), default_properties, namespace.uri_to_id
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Union

import nh3
import requests
from dependency_injector.wiring import Provide, inject
from werkzeug.exceptions import ServiceUnavailable

from sand.extension_interface.search import IEntitySearch, IOntologySearch, SearchResult
from sand.extensions.search.aggregated_search import AggregatedSearch
from sand.extensions.search.default_search import DefaultSearch
from sand.helpers.namespace import NamespaceService
from sand.models.ontology import OntClassAR, OntPropertyAR


def extended_wikidata_search() -> Union[IEntitySearch, IOntologySearch]:
    """extended version of wikidata search by aggregating default search"""
    search = AggregatedSearch()
    search.add(DefaultSearch.create())
    search.add(WikidataSearch())
    return search


class WikidataSearch(IEntitySearch, IOntologySearch):
    @inject
    def __init__(
        self,
        namespace: NamespaceService = Provide["namespace"],
        ontclass_ar: OntClassAR = Provide["classes"],
        ontprop_ar: OntPropertyAR = Provide["properties"],
    ):
        self.wikidata_url = "https://www.wikidata.org/w/api.php"
        self.PARAMS = {
            "action": "query",
            "format": "json",
            "list": "search",
            "srsearch": "",
            "utf8": "",
            "srnamespace": 0,
            "srlimit": 10,
            "srprop": "snippet|titlesnippet",
        }
        self.namespace = namespace
        self.ont_class_ar = ontclass_ar
        self.ont_prop_ar = ontprop_ar

    def get_class_search_params(self, search_text: str) -> Dict:
        """Updates class search parameters for wikidata API"""
        class_params = self.PARAMS.copy()
        class_params["srnamespace"] = 0
        class_params["srsearch"] = f"haswbstatement:P279 inlabel:{search_text}"
        return class_params

    def get_entity_search_params(self, search_text: str) -> Dict:
        """Updates entity search parameters for wikidata API"""
        entity_params = self.PARAMS.copy()
        entity_params["srnamespace"] = 0
        entity_params["srsearch"] = search_text
        return entity_params

    def get_props_search_params(self, search_text: str) -> Dict:
        """Updates property search parameters for wikidata API"""
        props_params = self.PARAMS.copy()
        props_params["srnamespace"] = 120
        props_params["srsearch"] = search_text
        return props_params

    def find_class_by_name(self, search_text: str) -> List[SearchResult]:
        """Uses Wikidata API to search for classes using their name/text."""
        if len(search_text.strip()) == 0:
            return []

        request_params = self.get_class_search_params(search_text)
        api_data = requests.get(self.wikidata_url, request_params)
        payload_results = []

        api_data_json = api_data.json()

        if "error" not in api_data.json():
            search_results = api_data.json()["query"]["search"]
        else:
            raise ServiceUnavailable(description=api_data_json["error"]["info"])

        for search_result in search_results:
            clsid = search_result["title"]
            if clsid not in self.ont_class_ar:
                continue
            cls = self.ont_class_ar[clsid]
            item = SearchResult(
                label=cls.label,
                id=clsid,
                description=cls.description,
                uri=self.namespace.id_to_uri(clsid),
            )
            payload_results.append(item)
        return payload_results

    def find_entity_by_name(self, search_text: str) -> Any:
        """Uses Wikidata API to search for entities using their name/text."""
        if len(search_text.strip()) == 0:
            return []

        request_params = self.get_entity_search_params(search_text)
        api_data = requests.get(self.wikidata_url, request_params)
        payload_results = []

        api_data_json = api_data.json()

        if "error" not in api_data_json:
            search_results = api_data_json["query"]["search"]
        else:
            raise ServiceUnavailable(description=api_data_json["error"]["info"])

        for search_result in search_results:
            item = SearchResult(
                label=nh3.clean(search_result["titlesnippet"], tags=set()),
                id=search_result["title"],
                description=nh3.clean(search_result["snippet"], tags=set()),
                uri=self.namespace.id_to_uri(search_result["title"]),
            )
            payload_results.append(item)
        return payload_results

    def find_props_by_name(self, search_text: str) -> List[SearchResult]:
        """Uses Wikidata API to search for properties using their name/text."""
        if len(search_text.strip()) == 0:
            return []

        request_params = self.get_props_search_params(search_text)
        api_data = requests.get(self.wikidata_url, request_params)
        payload_results = []

        api_data_json = api_data.json()

        if "error" not in api_data.json():
            search_results = api_data.json()["query"]["search"]
        else:
            raise ServiceUnavailable(description=api_data_json["error"]["info"])

        for search_result in search_results:
            propid = search_result["title"].split(":")[1]
            if propid not in self.ont_prop_ar:
                continue
            item = SearchResult(
                label=nh3.clean(search_result["titlesnippet"], tags=set()),
                id=propid,
                description=nh3.clean(search_result["snippet"], tags=set()),
                uri=self.namespace.id_to_uri(propid),
            )
            payload_results.append(item)
        return payload_results



from __future__ import annotations

import functools
from contextlib import contextmanager
from inspect import signature

from dependency_injector import containers
from dependency_injector.wiring import Provide, ProvidersMap
from loguru import logger

CURRENT_APP_CONTAINER: containers.Container | None = None
CURRENT_APP_PROVIDER_MAP: ProvidersMap | None = None


@contextmanager
def use_auto_inject(container: containers.Container):
    global CURRENT_APP_PROVIDER_MAP, CURRENT_APP_CONTAINER

    CURRENT_APP_CONTAINER = container
    CURRENT_APP_PROVIDER_MAP = ProvidersMap(container)

    yield None

    CURRENT_APP_CONTAINER = None
    CURRENT_APP_PROVIDER_MAP = None


def autoinject(func):
    inject_args: list[tuple[int, str, Provide]] = []

    for i, (name, param) in enumerate(signature(func).parameters.items()):
        if isinstance(param.default, Provide):
            inject_args.append((i, name, param.default))

    if len(inject_args) == 0:
        logger.error(
            "Function {} does not have any arguments with default values of Provide[X]"
            "Hence, no need to inject anything.",
            func.__qualname__,
        )
        raise Exception("no dependencies to inject")

    @functools.wraps(func)
    def fn(*args, **kwargs):
        if CURRENT_APP_PROVIDER_MAP is None:
            raise Exception(
                "Auto injection must be initialized using `use_autoinject` first"
            )

        for argidx, argname, provider in inject_args:
            if argidx < len(args) or argname in kwargs:
                continue
            prov = CURRENT_APP_PROVIDER_MAP.resolve_provider(
                provider.provider, provider.modifier
            )
            assert prov is not None
            kwargs[argname] = prov()
        return func(*args, **kwargs)

    return fn



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional

from sm.outputs.semantic_model import SemanticModel

from sand.models.table import Link, Table, TableRow


class IAssistant:
    def predict(
        self, table: Table, rows: list[TableRow]
    ) -> tuple[Optional[SemanticModel], Optional[list[TableRow]]]:
        """Predict semantic model and link entities"""
        raise NotImplementedError()

    def predict_entities(
        self, table: Table, rows: list[TableRow], row: int, col: int
    ) -> list[Link]:
        """Predict entities for a cell"""
        raise NotImplementedError()



from dataclasses import asdict

from dependency_injector.wiring import Provide, inject
from flask import Blueprint
from flask.json import jsonify

from sand.config import AppConfig

setting_bp = Blueprint("settings", "settings")


@setting_bp.route(f"/{setting_bp.name}", methods=["GET"])
@inject
def get_settings(appcfg: AppConfig = Provide["appcfg"]):
    return jsonify(
        {
            "entity": {
                "nil": asdict(appcfg.entity.nil),
                "instanceof": appcfg.entity.instanceof,
            },
            "semantic_model": {
                "identifiers": appcfg.semantic_model.identifiers,
                "statements": appcfg.semantic_model.statements,
            },
        }
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Mapping, Union, cast

from dependency_injector.wiring import Provide, inject
from sm.misc.funcs import import_func
from sm.namespaces.namespace import KnowledgeGraphNamespace

from sand.config import AppConfig
from sand.extension_interface.export import IExport
from sand.helpers.service_provider import MultiServiceProvider

if TYPE_CHECKING:
    from sand.models.entity import Entity
    from sand.models.ontology import OntClass, OntProperty


class NamespaceService:
    @inject
    def __init__(
        self,
        appcfg: AppConfig = Provide["appcfg"],
        default_entities: Mapping[str, Entity] = Provide["default_entities"],
        default_classes: Mapping[str, OntClass] = Provide["default_classes"],
        default_properties: Mapping[str, OntProperty] = Provide["default_properties"],
    ):
        self.appcfg = appcfg
        self.kgns: KnowledgeGraphNamespace = import_func(appcfg.kgns)()

        self.kgns_prefixes = self.kgns.prefix2ns.copy()

        self.default_entities = default_entities
        self.default_classes = default_classes
        self.default_properties = default_properties

        self.uri2resource = {}
        self.id2resource = {}

        for odict in [
            self.default_entities,
            self.default_classes,
            self.default_properties,
        ]:
            for obj in cast(
                Mapping[str, Union["Entity", "OntClass", "OntProperty"]], odict
            ).values():
                assert obj.uri not in self.uri2resource
                self.uri2resource[obj.uri] = obj
                self.id2resource[obj.id] = obj

    def uri_to_id(self, uri: str):
        """Convert an URI to the corresponding ID"""
        if uri in self.uri2resource:
            return self.uri2resource[uri].id
        return self.kgns.uri_to_id(uri)

    def id_to_uri(self, id: str):
        """Convert an ID to the corresponding URI"""
        if id in self.id2resource:
            return self.id2resource[id].uri
        return self.kgns.id_to_uri(id)

    def has_encrypted_name(self, uri: str):
        """Check if the given uri of an entity/class/property in the KG has encrypted name such as QXXX so we need to add
        label to make it readable.
        """
        return self.kgns.has_encrypted_name(uri)



from __future__ import annotations

import os
from contextlib import contextmanager
from pathlib import Path
from typing import Optional

from dependency_injector import containers, providers
from sm.misc.funcs import import_func

from sand.config import PACKAGE_DIR, AppConfig
from sand.helpers.namespace import NamespaceService
from sand.helpers.service_provider import MultiServiceProvider, get_service
from sand.models.entity import EntityAR, get_default_entities
from sand.models.ontology import (
    OntClassAR,
    OntPropertyAR,
    get_default_classes,
    get_default_properties,
)


class AppContainer(containers.DeclarativeContainer):
    basecfg = providers.Configuration()

    # app configuration
    appcfg = providers.Singleton(AppConfig.from_yaml, basecfg.config_file)

    # namespace service
    namespace = providers.Singleton(NamespaceService)

    # KG databases
    entities = providers.Singleton(EntityAR.init)
    classes = providers.Singleton(OntClassAR.init)
    properties = providers.Singleton(OntPropertyAR.init)

    default_entities = providers.Singleton(get_default_entities)
    default_classes = providers.Singleton(get_default_classes)
    default_properties = providers.Singleton(get_default_properties)

    assistant = providers.Singleton(MultiServiceProvider, cfg=appcfg.provided.assistant)
    export = providers.Singleton(MultiServiceProvider, cfg=appcfg.provided.export)
    entity_search = providers.Singleton(get_service, path=appcfg.provided.search.entity)
    ontology_search = providers.Singleton(
        get_service, path=appcfg.provided.search.ontology
    )


@contextmanager
def use_container(config_file: Optional[Path | str] = None):
    container = AppContainer()
    container.basecfg.from_dict(
        {
            "config_file": str(
                config_file or os.path.join(PACKAGE_DIR, "config.default.yml")
            )
        }
    )

    modules = []
    export = container.export()
    for func in export.cfg.funcs.values():
        modules.append(func.rsplit(".", 1)[0])

    try:
        container.wire(
            packages=[
                "sand.controllers",
                "sand.models",
                "sand.extensions.assistants",
                "sand.extensions.search",
                "sand_drepr.main",
                "sand.deserializer",
                "sand.serializer",
                "sand.helpers",
                "sand.app",
                "sand.commands.load",
            ],
            modules=modules,
        )
        yield container
    finally:
        container.unwire()



from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum

import sm.outputs.semantic_model as O

from sand.models.table import Table, TableRow


class OutputFormat(str, Enum):
    TTL = "ttl"


class IExport(ABC):
    """
    Export interface class to export data to different data formats
    """

    @abstractmethod
    def export_data_model(self, table: Table, sm: O.SemanticModel) -> dict[str, str]:
        """Export data model in multiple formats"""
        pass

    @abstractmethod
    def export_extra_resources(
        self, table: Table, rows: list[TableRow], sm: O.SemanticModel
    ) -> dict[str, str]:
        """Export extra resources generated automatically by the system"""
        pass

    @abstractmethod
    def export_data(
        self,
        table: Table,
        rows: list[TableRow],
        sm: O.SemanticModel,
        output_format: OutputFormat,
    ):
        """Export relational data"""
        pass



import copy
import subprocess
from pathlib import Path
from typing import Dict, List, Tuple

import serde.prelude as serde
import sm.outputs.semantic_model as O
from rdflib import RDFS
from sm.namespaces.wikidata import WikidataNamespace

from sand.controllers.assistant import IAssistant
from sand.models.base import init_db
from sand.models.table import CandidateEntity, Link, Table, TableRow


class MTabAssistant(IAssistant):
    def __init__(
        self,
    ):
        self.cache_dir = Path(f"/tmp/mtab")
        self.cache_dir.mkdir(exist_ok=True)
        self.wdns = WikidataNamespace.create()
        self.id2label = {}

    def predict(self, table: Table, rows: List[TableRow]):
        content = [table.columns]
        for row in rows:
            content.append([str(x) for x in row.row])

        cwd = self.cache_dir / f"p{table.project.id}-t:{table.id}-{table.name}"
        cwd.mkdir(exist_ok=True)
        infile = cwd / "infile.csv"
        outfile = cwd / "outfile.json"

        rerun = True
        if infile.exists() and outfile.exists():
            serde.csv.ser(content, str(infile) + ".tmp")
            if serde.textline.deser(infile) == serde.textline.deser(
                str(infile) + ".tmp"
            ):
                rerun = False

        if rerun:
            serde.csv.ser(content, infile)
            subprocess.check_call(
                [
                    "curl",
                    "-X",
                    "POST",
                    "-F",
                    f"file=@{str(infile)}",
                    "https://mtab.app/api/v1/mtab?limit=1000",
                    "-o",
                    outfile,
                ]
            )

        record = serde.json.deser(outfile)
        assert record["status"] == "Success"

        semantic = record["tables"][0]["semantic"]

        cpa, cta = [], {}
        for item in semantic["cpa"]:
            source, target = item["target"]
            # assert len(item["annotation"]) == 1
            uri = item["annotation"][0]["wikidata"]
            label = item["annotation"][0]["label"]
            assert uri.startswith("http://www.wikidata.org/prop/direct/")
            pid = uri.replace("http://www.wikidata.org/prop/direct/", "")
            self.id2label[pid] = label
            cpa.append((source, target, pid))

        for item in semantic["cta"]:
            # assert len(item["annotation"]) == 1
            uri = item["annotation"][0]["wikidata"]
            label = item["annotation"][0]["label"]
            assert uri.startswith("http://www.wikidata.org/entity/")
            qid = uri.replace("http://www.wikidata.org/entity/", "")
            cta[item["target"]] = qid
            self.id2label[qid] = label

        pred_sm = self.create_sm_from_cta_cpa(table.columns, cpa, cta)

        # clean previous data -- temporary solution for now before we tag each result with
        # individual method
        rows = copy.deepcopy(rows)
        for row in rows:
            row.links = {}

        for ann in semantic["cea"]:
            ri, ci = ann["target"]
            ri = ri - 1
            if ri == -1:
                continue

            uri = ann["annotation"]["wikidata"]
            if uri.startswith("http://www.wikidata.org/entity/"):
                entid = uri.replace("http://www.wikidata.org/entity/", "")
            elif uri.startswith("http://www.wikidata.org/prop/direct/"):
                entid = uri.replace("http://www.wikidata.org/prop/direct/", "")
            else:
                raise ValueError(uri)

            if str(ci) not in rows[ri].links:
                rows[ri].links[str(ci)] = [
                    Link(
                        start=0,
                        end=len(str(rows[ri].row[ci])),
                        url=None,
                        entity_id=None,
                        candidate_entities=[],
                    )
                ]
            rows[ri].links[str(ci)][0].candidate_entities.append(
                CandidateEntity(entity_id=entid, probability=1.0)
            )

        return pred_sm, rows

    def create_sm_from_cta_cpa(
        self,
        columns: List[str],
        rels: List[Tuple[int, int, str]],
        types: Dict[int, str],
    ) -> O.SemanticModel:
        """Create a semantic model from a list of columns, CPA, and CTA.

        Args:
            columns: list of column names
            rels: list of (source column index, target column index, prop id not url)
            types: a mapping from column index to WDEntity id (not url)
        """
        sm = O.SemanticModel()

        # create class nodes first
        classmap = {}

        for cid, qnode_id in types.items():
            dnode = O.DataNode(
                col_index=cid,
                label=columns[cid],
            )

            # somehow, they may end-up predict multiple classes, we need to select one
            if qnode_id.find(" ") != -1:
                qnode_id = qnode_id.split(" ")[0]
            curl = self.wdns.id_to_uri(qnode_id)

            try:
                cnode_label = f"{self.id2label[qnode_id]} ({qnode_id})"
            except KeyError:
                cnode_label = self.wdns.get_rel_uri(self.wdns.id_to_uri(qnode_id))
            cnode = O.ClassNode(
                abs_uri=curl,
                rel_uri=self.wdns.get_rel_uri(self.wdns.id_to_uri(qnode_id)),
                readable_label=cnode_label,
            )
            sm.add_node(dnode)
            sm.add_node(cnode)
            sm.add_edge(
                O.Edge(
                    source=cnode.id,
                    target=dnode.id,
                    abs_uri=str(RDFS.label),
                    rel_uri="rdfs:label",
                    readable_label="rdfs:label",
                )
            )

            classmap[dnode.id] = cnode.id

        # now complete the relationships
        for source_cid, target_cid, prop in rels:
            if sm.has_data_node(source_cid):
                source = sm.get_data_node(source_cid)
                # if source node is an NE column, we need to use a class id instead
                if source.id in classmap:
                    source = sm.get_node(classmap[source.id])
            else:
                # create a data node
                source = O.DataNode(
                    col_index=source_cid,
                    label=columns[source_cid],
                )
                sm.add_node(source)

            if sm.has_data_node(target_cid):
                # if target node is an NE column, we need to use a class id instead
                target = sm.get_data_node(target_cid)
                if target.id in classmap:
                    target = sm.get_node(classmap[target.id])
            else:
                # create a data node
                target = O.DataNode(col_index=target_cid, label=columns[target_cid])
                sm.add_node(target)

            sm.add_edge(
                O.Edge(
                    source=source.id,
                    target=target.id,
                    abs_uri=(tmp_abs_uri := self.wdns.id_to_uri(prop)),
                    rel_uri=self.wdns.get_rel_uri(tmp_abs_uri),
                    readable_label=f"{self.id2label[prop]} ({prop})",
                )
            )

        # complete missing data nodes
        for ci, col in enumerate(columns):
            if sm.has_data_node(ci):
                continue
            sm.add_node(O.DataNode(col_index=ci, label=col))
        return sm


if __name__ == "__main__":
    init_db("/Users/rook/workspace/sm-dev/data/home/databases/smc.db")
    table = Table.get_by_id(501)
    rows = list(
        TableRow.select().where(TableRow.table == table).order_by(TableRow.index)
    )

    assistant = MTabAssistant()
    resp = assistant.predict(table, rows)
    print(resp)
